#!/usr/bin/python3

import sys, os

FORMAT_COLORS = \
{
    "GREEN": "\033[32m",
    "RED": "\033[31m",
    "YELLOW": "\033[33m",
    "BLUE": "\033[94m",
    "PURPLE": "\033[35m"
}

FORMAT_RESET = "\033[0m"
FORMAT_OUTPUT = sys.stdout.isatty() # Only format if outputting to a terminal.

# Wrap a given value in a file descriptor. Permits calling cprint
# with file=a file descriptor.
class _FDWrap:
    def __init__(self, fd):
        self.fd = fd
    def write(self, txt):
        return os.write(self.fd, bytes(txt, 'utf-8'))
    def read(self):
        result = ''
        part = os.read(self.fd, 1).decode('utf-8')

        while part != '':
            result += part or ''
            part = os.read(self.fd, 1).decode('utf-8')
        
        return result
    def flush(self):
        pass

# If file is None, default to stdout. If a number, 
# wrap it in a file-like object.
def wrapFile(fileOrFd):
    if fileOrFd is None:
        return sys.stdout
    elif type(fileOrFd) == int:
        return _FDWrap(fileOrFd)
    return fileOrFd

def cprint(text, color=None, file=sys.stdout):
    file = wrapFile(file)

    if color in FORMAT_COLORS and FORMAT_OUTPUT:
        print(FORMAT_COLORS[color] + str(text) + FORMAT_RESET, end='', flush=True, file=file)
    elif type(color) == str and FORMAT_OUTPUT:
        print(color + str(text) + FORMAT_RESET, end='', flush=True, file=file)
    else:
        print(text, end='', flush=True, file=file)



#!/usr/bin/python3

# See https://danishpraka.sh/2018/09/27/shell-in-python.html (Accessed Aug 22)

import cmd, os, sys, shutil, pathlib, re

from almost_make.utils.printUtil import cprint, FORMAT_COLORS
import almost_make.utils.printUtil as printer
from almost_make.utils.argsUtil import parseArgs
from almost_make.version import printVersion
import almost_make.utils.shellUtil.runner as runner
import almost_make.utils.shellUtil.escapeParser as escapeParser

SHELL_NAME = "AlmostMake's built-in shell"

QUICK_HELP = \
{
    "cd": """Usage: cd [directory]
Change the current working directory to [directory].
""",

    "exit": """Usage: exit [status]
Exit the current script/shell with status [status].
If [status] is not provided, exit with code zero.
""",

    "ls": """Usage: ls [options] [directories]
Print the contents of each directory in [directories] 
or the current working directory.
[options] can contain:
 -a, --all                   List all files, including '.' and '..'
 -1, --one-per-line          Each file on its own line. This is assumed when output does not appear to be to a terminal.
 -Q, --quote-name            Surround the name of each file with double-quotes.
 -m, --comma-separated-list  List files on a single line, as a comma-separated list.
 -f, --unformatted           No color output, do not sort.
 --color                     Force color output. Do not assume --one-per-line when output does not seem to be to a terminal.
 """,

    "pwd": """Usage: pwd
Print the current working directory's absolute path.""",

    "echo": """Usage: echo [options] [text]
Send [text] to standard output.
[options] can contain:
 -n\t\t Do not print a trailing newline.
 -e\t\t Evaluate escape characters (e.g. echo -ne \\n outputs a single newline).
""",

    "touch": """Usage: touch [options] [files...]
Update the access and modification times for each file in files.
Options:
 -c, --no-create  Do not create the file if it does not exist.
""",

    "cat": """Usage: cat [options] [files...]
Print the contents of each file in [files]. If a file is -, print standard input, instead.
Options:
 -n, --number    Number each line.
 -T, --show-tabs Replace all tab characters with ^T.
 -E, --show-ends Print a $ at the end of each line.
""",
    "grep": """Usage: grep [options] PATTERN
Search for PATTERN in standard input.
Options:
 -F, --fixed-strings   Treat PATTERN as a fixed string, rather than a regular expression.
 -i, --ignore-case     Do case-insensitive matching.
 -v, --invert-match    Count and print (print depends on other options) lines that do not match PATTERN.
 -c, --count           Print the count of matching lines, rather than the contents of the lines.
 -q, --quiet           Limit output.
 -x, --line-regexp     PATTERN must match each line in its entirety.
 -o, --only-matching   Only output portions of the line that match PATTERN. Ignored if --line-regexp or -x are given.
 -n, --line-number     Prefix each printed-line with a line number.
 --no-color            Force uncolorized output.
"""
}

def filterArgs(args, minimumLength, stdout):
    if len(args) < minimumLength or "--help" in args:
        cprint(QUICK_HELP[args[0]] + '\n', file=stdout)
        cprint("This command is built into %s.\n" % SHELL_NAME)
        return False
    elif '--version' in args:
        printVersion(printer.wrapFile(stdout))
        return False
    return True

CUSTOM_COMMANDS = \
{
    "exit": lambda args, flags, stdin, stdout, stderr, state: filterArgs(args, 1, stdout) and sys.exit((len(args) > 1 and args[1]) or 0)
}

LS_DIRECTORY_COLOR = FORMAT_COLORS['BLUE']
LS_LINK_COLOR = FORMAT_COLORS['BLUE']
LS_FILE_COLOR = None

def customLs(args, stdin, stdout, stderr, state):
    dirs = [ os.path.abspath(state.cwd or '.') ]
    
    args = parseArgs(args, 
    {
        'a': 'all',
        'f': 'unformatted',
        '1': 'one-per-line',
        'Q': 'quote-name',
        'm': 'comma-separated-list'
    },
    strictlyFlags =
    {
        'all', 'unformatted', 'one-per-line',
        'quote-name', 'comma-separated-list',
        'color'
    })

    if 'default' in args and len(args['default']) > 0:
        dirs = list(map(os.path.abspath, args['default']))
    
    def noteEntry(name, color, isLast = False):
        decolorized = False # If we decolorize, do other formatting...

        # If given a file descriptor (not default output),
        # we probably aren't sending output to a terminal. As such,
        # remove coloring. --color overrides this.
        if stdout != None and not 'color' in args:
            decolorized = True
            color = None

        sep = '  '

        if (not 'all' in args and not 'unformatted' in args) and name.startswith('.'):
            return

        if 'one-per-line' in args or decolorized:
            sep = not isLast and '\n' or ''

        if 'quote-name' in args:
            name = runner.quote(name, '"')

        if not isLast and 'comma-separated-list' in args:
            sep = ', '

            if 'one-per-line' in args or decolorized:
                sep = ',\n'

        if 'unformatted' in args:
            cprint(name + sep, file=stdout)
        else:
            cprint(name + sep, color, file=stdout)

    if not 'unformatted' in args:
        dirs.sort()

    isFirst = True

    for directory in dirs:
        if len(dirs) > 1:
            if not isFirst:
                cprint("\n", file=stdout)
            
            isFirst = False
            cprint("%s:\n" % directory, file=stdout)
        
        fileList = []

        with os.scandir(directory) as files:
            if not 'unformatted' in args:
                fileList = sorted(files, key=lambda entry: entry.name)
            else:
                fileList = list(files)

        noteEntry('.', LS_DIRECTORY_COLOR)
        noteEntry('..', LS_DIRECTORY_COLOR, isLast = (len(fileList) == 0))

        if len(fileList) != 0:
            for entry in fileList[:-1]:
                noteEntry(entry.name, entry.is_dir() and LS_DIRECTORY_COLOR or LS_FILE_COLOR)
            noteEntry(fileList[-1].name, fileList[-1].is_dir() and LS_DIRECTORY_COLOR or LS_FILE_COLOR, isLast = True)
        
        cprint("\n", file=stdout)

def customPwd(args, stdin, stdout, stderr, state):
    cprint(os.path.abspath(state.cwd or '.') + '\n', file=stdout)

def customTouch(args, stdin, stdout, stderr, state):
    args = parseArgs(args, 
    {
        'c': 'no-create'
    }, 
    strictlyFlags=
    {
        'no-create'
    })

    cwd = pathlib.Path(state.cwd or '.')
    touchedCount = 0

    for path in args['default']:
        path = cwd.joinpath(path)

        if path.is_file() or not 'no-create' in args:
            path.touch()
            touchedCount += 1
    
    return touchedCount != 0

def customCd(args, stdin, stdout, stderr, state):
    oldCwd = state.cwd

    if not state.cwd:
        state.cwd = os.path.abspath(os.path.expanduser(args[1]))
    else:
        state.cwd = os.path.abspath(os.path.expanduser(os.path.join(state.cwd, os.path.expanduser(args[1]))))
    
    result = True

    if not os.path.exists(state.cwd):
        cprint("cd: " + str(state.cwd) + ": No such file or directory\n", file=stderr)
        state.cwd = oldCwd
        result = False
    
    return result

def customEcho(args, stdin, stdout, stderr, state):
    if len(args) == 1:
        return 0
    
    doEscapes = False
    doNewlines = True
    firstArg = args[1].strip()
    
    if firstArg.startswith("-") and len(firstArg) <= 3:
        doEscapes = 'e' in firstArg
        doNewlines = not 'n' in firstArg
        
        args = args[2:]
        
        if len(args) > 0 and args[0].startswith('-') and len(args[0]) <= 2:
            if 'e' in args[0]:
                doEscapes = True
                trimArg = True
            if 'n' in args[0]:
                doNewlines = False
                trimArg = True
            if trimArg:
                args = args[1:]
    else:
        args = args[1:]
    
    
    if len(args) < 1:
        return 0
    
    printEnd = '\n'
    toPrint = " ".join(args)
    
    if not doNewlines:
        printEnd = ''
    
    if doEscapes:
        toPrint = escapeParser.parseEscapes(toPrint)
    
    cprint(toPrint + printEnd, file=stdout)

def customCat(args, stdin, stdout, stderr, state):
    args = parseArgs(args,
    {
        'n': 'number',
        'T': 'show-tabs',
        'E': 'show-ends'
    },
    strictlyFlags=
    {
        'number',
        'show-tabs',
        'show-ends'
    })

    lineNu = 0

    def logLine(line):
        if 'number' in args:
            cprint("%s\t" % str(lineNu), file=stdout)
        
        if 'show-tabs' in args:
            line = re.sub(r'[\t]', '^T', line)
        
        end = 'show-ends' in args and '$' or ''
        cprint(str(line) + end + "\n", file=stdout)

    stdin = printer.wrapFile(stdin)
    
    for arg in args['default']:
        if arg == '-':
            lines = stdin.read().split('\n')

            if len(lines) > 0 and lines[-1] == '':
                lines = lines[:-1]

            for line in lines:
                lineNu += 1
                logLine(line)
        else:
            filename = os.path.join(state.cwd or '.', arg)

            if not os.path.exists(filename):
                cprint("File %s does not exist.\n", printer.FORMAT_COLORS["red"], file=stderr)
                return False
            if not os.path.isfile(filename):
                cprint("Path %s is not a file.\n", printer.FORMAT_COLORS["red"], file=stderr)
                return False
            
            with open(filename, 'r') as file:
                lines = file.read().split('\n')

                if len(lines) > 0 and lines[-1] == '':
                    lines = lines[:-1]

                for line in lines:
                    lineNu += 1
                    logLine(line)

def customGrep(args, stdin, stdout, stderr, state):
    args = parseArgs(args,
    {
        'F': 'fixed-strings',
        'i': 'ignore-case',
        'v': 'invert-match',
        'c': 'count',
        'q': 'quiet',
        'x': 'line-regexp',
        'o': 'only-matching',
        'n': 'line-number'
    },
    strictlyFlags=
    {
        'fixed-strings', 'ignore-case', 'invert-match',
        'count', 'quiet', 'line-regexp', 'only-matching',
        'line-number', 'no-color'
    })

    if len(args['default']) > 1:
        cprint("[Files...] is currently unsupported. Input should be given via stdin.\n", printer.FORMAT_COLORS['red'], file=stderr)
        return False
    
    patterns = []

    if len(args['default']) > 0:
        patternList = args['default'][0].split('\n')

        for part in patternList:
            flags = 0

            if 'fixed-strings' in args:
                part = re.escape(part)
            if 'ignore-case' in args:
                flags = flags | re.IGNORECASE

            patterns.append(re.compile(part, flags))
    
    def matchesLine(line):
        matches = []
        for pattern in patterns:
            matchInfo = None

            if not 'line-regexp' in args:
                matchInfo = pattern.search(line)
            else:
                matchInfo = pattern.fullmatch(line)

            if matchInfo != None:
                matches.append(matchInfo.span())
        return matches
    
    stdin = printer.wrapFile(stdin)
    lines = stdin.read().split('\n')

    # Input is from stdin... There is often a trailing newline...
    if len(lines) > 0 and lines[-1] == '\n':
        lines = lines[:-1]

    lineNumber = 0
    matchCount = 0

    for line in lines:
        lineNumber += 1
        matches = matchesLine(line)

        # Negates (len(matches) > 0) if 'invert-match' in args.
        if (len(matches) > 0) == (not 'invert-match' in args):
            matchCount += 1 # Count the number of **lines**

            if not 'count' in args and not 'quiet' in args:
                if 'line-number' in args:
                    cprint(str(lineNumber) + '\t', file=stdout)

                if not 'only-matching' in args or 'invert-match' in args:
                    if stdout != None or 'no-color' in args: # Don't colorize output when output isn't the default...
                        cprint(line + '\n', file=stdout)
                    else: # Otherwise, colorize output.
                        startIndexes = set()
                        stopIndexes = set()

                        for start,stop in matches:
                            startIndexes.add(start)
                            stopIndexes.add(stop)
                        
                        buff = ''
                        inRange = False

                        for i in range(0, len(line)):
                            if i in startIndexes and not inRange:
                                inRange = True
                            elif i in stopIndexes and inRange:
                                cprint(buff, printer.FORMAT_COLORS['PURPLE'], file=stdout)
                                inRange = False
                                buff = ''
                            if inRange:
                                buff += line[i]
                            else:
                                cprint(line[i], file=stdout)
                        
                        cprint(buff, printer.FORMAT_COLORS['PURPLE'], file=stdout)
                        cprint('\n', file=stdout)
                else:
                    for start,stop in matches:
                        cprint(line[start:stop] + '\n', file=stdout)
    
    if 'count' in args:
        cprint(str(matchCount) + '\n', file=stdout)
    
    return matchCount > 0

# Get a set of custom commands that can be used.
def getCustomCommands(macros):
    result = {}
    
    for key in CUSTOM_COMMANDS:
        result[key] = CUSTOM_COMMANDS[key]

    def addCustomCommand(alias, minArgs, fn):
        result[alias] = lambda args, flags, stdin, stdout, stderr, state: filterArgs(args, minArgs, stdout) and fn(args, stdin, stdout, stderr, state)
    
    addCustomCommand("cd", 2, customCd)

    if "_CUSTOM_BASE_COMMANDS" in macros:
        addCustomCommand("ls", 1, customLs)
        result["dir"] = result["ls"]
        addCustomCommand("pwd", 1, customPwd)
        addCustomCommand("echo", 2, customEcho)
        addCustomCommand("touch", 2, customTouch)
        addCustomCommand("cat", 2, customCat)
        addCustomCommand("grep", 2, customGrep)
    
    return result

def evalScript(text, macroUtil, macros={}, defaultFlags = [], state=None):
    if not state:
        state = runner.ShellState()
    
    # Set appropriate macros.
    macros["PWD"] = state.cwd

    text, macros = macroUtil.expandAndDefineMacros(text, macros)
    return (runner.runCommand(text, getCustomCommands(macros), defaultFlags, state), macros)

if __name__ == "__main__":
    import almost_make.utils.macroUtil as macroUtility
    macroUtil = macroUtility.MacroUtil()

    def assertEql(a, b, message):
        if a != b:
            raise Exception("%s != %s (%s)" % (str(a), str(b), message))
    
    def assertNotEql(a, b, message):
        if a == b:
            raise Exception("%s == %s (%s)" % (str(a), str(b), message))

    testDir = os.path.dirname(__file__)

    if testDir != '':
        os.chdir(testDir)

    macros = \
    {
        "_CUSTOM_BASE_COMMANDS": True
    }

    result, _ = evalScript("ls | grep __init__.py", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test ls and grep in current directory for __init__.")

    result, _ = evalScript("ls -f | grep -F ..", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test ls with -f flag.")

    result, _ = evalScript("ls -f . | grep -F ..", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test ls with provided directory")

    result, _ = evalScript("ls -f ./ ../ | grep -F argsUtil.py", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test ls with provided directories (1 of 2)")

    result, _ = evalScript("ls -f ./ ../ | grep -F escapeParser.py", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test ls with provided directories (2 of 2)")

    result, _ = evalScript("echo -e 'F\\noo' | grep ^F$", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test echo's -e flag.")

    result, _ = evalScript("echo test | cat - | grep test", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test cat from stdin")

    result, _ = evalScript("echo -e 'test\\nfoo' | cat -n - | grep -F 2", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test -n flag for cat (line numbers)")

    result, _ = evalScript("echo -e 'test\\n\\tfoo' | cat -T - | grep \\^Tfoo", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test -T flag for cat (tab->^T)")

    result, _ = evalScript("echo -e 'test\\n\\tfoo' | cat -TE - | grep '\\^Tfoo\\$$'", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test -E flag for cat (ending $)")

    result, _ = evalScript("cat __init__.py | grep '__all__'", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test cat [filename].")

    result, _ = evalScript("echo nothing | grep 'should not find anything'", macroUtil, macros, defaultFlags=[])
    assertEql(result, 1, "Test grep failure.")

    result, _ = evalScript("echo nothing | grep 'o[th]+ing$'", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep simple regexp.")

    result, _ = evalScript("echo nothing | grep -x noth", macroUtil, macros, defaultFlags=[])
    assertEql(result, 1, "Test grep full-line match failure.")

    result, _ = evalScript("echo '' | grep -x .*", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test (very simple) grep full-line match success.")

    result, _ = evalScript("echo nothing | grep -Fx nothing", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep full-line match success (with -F).")

    result, _ = evalScript("echo nothing | grep -x nothing", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep full-line match success.")

    result, _ = evalScript("echo -e 'a\\nbcd\\nefg' | grep -Fv 'nothing'", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep invert-match success.")

    result, _ = evalScript("echo textthenfood | grep -o foo | grep -Fx foo", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep only-match success.")

    result, _ = evalScript("echo textthenfood | grep -c foo | grep -Fx 1", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep count (1).")

    result, _ = evalScript("echo -e 'textthenfood\\nfoo' | grep -c foo | grep -Fx 2", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep count (2).")

    result, _ = evalScript("echo -e 'textthenfood\\nfoo' | grep -n foo | cat -T - | grep -Fx 2^Tfoo", macroUtil, macros, defaultFlags=[])
    assertEql(result, 0, "Test grep line number (2).")


#!/usr/bin/python3
import sys, os.path
from almost_make.utils.printUtil import *
import almost_make.utils.makeUtil as makeUtility
import almost_make.utils.macroUtil as macroUtility
from almost_make.utils.argsUtil import *
from almost_make import version

ARGUMENT_MAPPINGS = \
{
    'h': "help",
    'k': 'keep-going',
    'p': 'print-expanded',
    'n': 'just-print',
    'f': 'file',
    'C': 'directory',
    's': 'silent',
    'b': 'built-in-shell',
    'j': 'jobs'
}

# These are flags, so don't associate values with them...
JUST_FLAGS = \
{
    'help', 'keep-going', 'print-expanded', 'just-print', 'silent', 'built-in-shell'
}

# Don't save these when we recurse...
NO_SAVE_ARGS = \
{
    'C', 'directory',
    'f', 'file',
    'default',
    'h', 'help', 
    'version'
}

def printHelp():
    cprint("Help: \n", FORMAT_COLORS['YELLOW'])
    cprint(" Summary: ", FORMAT_COLORS['YELLOW'])
    print("Satisfy dependencies of a target in a makefile. This parser is not quite POSIX compliant, but should be able to parse simple makefiles.")
    cprint(" Usage: almake [targets...] [options]\n", FORMAT_COLORS['YELLOW'])
    print("  where each target in targets is a valid target and options include:")
    cprint("    -h, --help", FORMAT_COLORS['GREEN'])
    print("\t Print this message.")
    cprint("    --version", FORMAT_COLORS['GREEN'])
    print("\t Print version and licensing information.")
    cprint("    --file", FORMAT_COLORS['GREEN'])
    print("\t File to parse (default is Makefile).")
    cprint("    -k", FORMAT_COLORS['GREEN'])
    print("\t\t Keep going if errors are encountered.")
    cprint("    -n, --just-print", FORMAT_COLORS['GREEN'])
    print("\t Just print commands to be run, without evaluating (print commands, don't send them to the shell). ")
    print("Be aware that $(shell ...) macros are still evaluated. This option only applies to individual commands.")
    cprint("    -p", FORMAT_COLORS['GREEN'])
    print("\t\t Rather than finding targets, print the makefile, with top-level targets expanded.")
    cprint("    -C dir", FORMAT_COLORS['GREEN'])
    print("\t Switch to directory, dir, before running make. ")
    cprint("    -j, --jobs", FORMAT_COLORS['GREEN'])
    print("\t Maximum number of jobs (e.g. almake -j 8). ")
    cprint("    -s, --silent", FORMAT_COLORS['GREEN'])
    print(" In most cases, don't print output.")
    cprint("    -b, --built-in-shell", FORMAT_COLORS['GREEN'])
    print("\t Use the built-in shell for commands in the makefile. This can also be enabled as follows:")
    cprint("   export ", FORMAT_COLORS['PURPLE'])
    print("_BUILTIN_SHELL ", end='')
    cprint(":= ", FORMAT_COLORS['YELLOW'])
    print("1 \t\t", end='')
    cprint("# Use the built-in shell instead of the system shell.", FORMAT_COLORS['GREEN'])
    print()
    cprint("   export", FORMAT_COLORS['PURPLE'])
    print(" _CUSTOM_BASE_COMMANDS ", end='')
    cprint(":= ", FORMAT_COLORS['YELLOW'])
    print("1 \t", end='')
    cprint("# Enable built-in overrides for several commands like ls, echo, and pwd.", FORMAT_COLORS['GREEN'])
    print()
    cprint("   export", FORMAT_COLORS['PURPLE'])
    print(" _SYSTEM_SHELL_PIPES ", end='')
    cprint(":= ", FORMAT_COLORS['YELLOW'])
    print("1 \t", end='')
    cprint("# Send commands that seem related to pipes (e.g. ls | less) directly to the system's shell. ", FORMAT_COLORS['GREEN'])
    print()
    cprint("Note: ", FORMAT_COLORS['PURPLE'])
    print("AlmostMake's built-in shell is currently very limited.")
    print()
    cprint("Note: ", FORMAT_COLORS['PURPLE'])
    print("Macro definitions that override those from the environment" +
" can be provided in addition to targets and options. For example,")
    cprint("    make target1 target2 target3 CC=gcc CFLAGS=-O3", FORMAT_COLORS['YELLOW'])
    print()
    print("should make target1, target2, and target3 with the " +
          "macros CC and CFLAGS by default set to gcc and -O3, respectively.")
    cprint("Note: ", FORMAT_COLORS['PURPLE'])
    print("Options can also be given to almake through the environment. " +
    	  "This is done through the MAKEFLAGS variable. For example, " +
    	  "setting MAKEFLAGS to --built-in-shell causes almake to " +
    	  "always use its built-in shell, rather than the system shell.")

# On commandline run...
def main(args=sys.argv):
    args = parseArgs(args, ARGUMENT_MAPPINGS, strictlyFlags=JUST_FLAGS)
    
    # Fill args from MAKEFLAGS (see https://www.gnu.org/software/make/manual/make.html#How-the-MAKE-Variable-Works)
    args = fillArgsFromEnv(args, "MAKEFLAGS", ARGUMENT_MAPPINGS) # Previously-defined args take precedence.
    saveArgsInEnv(args, "MAKEFLAGS", NO_SAVE_ARGS) # For recursive calls to make.
    
    if 'help' in args:
        printHelp()
    elif 'version' in args:
        version.printVersion()
    else:
        macroUtil = macroUtility.MacroUtil()
        makeUtil = makeUtility.MakeUtil()

        fileName = 'Makefile'
        targets = []
        
        defaultMacros = macroUtil.getDefaultMacros() # Fills with macros from environment, etc.
        overrideMacros = {}
        
        if 'directory' in args:
            try:
                os.chdir(args['directory'])
            except Exception as ex:
                print("Error changing directories: %s" % str(ex))
                sys.exit(1)
        
        # If we know the path to the python interpreter...
        if sys.executable:
            defaultMacros["MAKE"] = sys.executable + " " + os.path.abspath(__file__) 
                                #^ Use ourself, rather than another make implementation.

        if 'keep-going' in args:
            makeUtil.setStopOnError(False)
        
        if 'silent' in args:
            makeUtil.setSilent(True)

        if 'jobs' in args:
            jobs = 1
            try:
                jobs = int(args['jobs'])
            except ValueError as ex:
                makeUtil.errorUtil.reportError("Invalid argument to --jobs. This must be an integer.")
            makeUtil.setMaxJobs(jobs)

        if 'just-print' in args:
            makeUtil.setJustPrint(True)

        if 'file' in args:
            fileName = args['file']

        if len(args['default']) > 0:
            targets = [ ]
            
            # Split into targets and default macros.
            for arg in args['default']:
                assignmentIndex = arg.find("=")
                if assignmentIndex > 0:
                    key = arg[:assignmentIndex].strip() # e.g. VAR in VAR=33
                    val = arg[assignmentIndex+1:].strip() # e.g. 33 in VAR=33
                    overrideMacros[key] = val
                    defaultMacros[key] = val
                else:
                    targets.append(arg)
	
	    # Were we told to use the built-in shell?
        if 'built-in-shell' in args:
            overrideMacros["_BUILTIN_SHELL"] = "1"
            overrideMacros["_CUSTOM_BASE_COMMANDS"] = "1"
	
        if len(targets) == 0: # Select the default target, if no targets
            targets = ['']
        
        if not os.path.exists(fileName):
            cprint("The file with name \"%s\" was not found!\n" % fileName, FORMAT_COLORS['RED'])
            print("Please check your spelling.")
            sys.exit(1)

        fileObj = open(fileName, 'r')
        fileContents = fileObj.read()
        fileObj.close()
        
        if not 'print-expanded' in args:
            # Run for each target.
            for target in targets:
                makeUtil.runMakefile(fileContents, target, defaultMacros, overrideMacros)
        else:
            contents, macros = macroUtil.expandMacros(fileContents, defaultMacros)
            print(contents)

if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Macro parsing utilities.

import re, os
import almost_make.utils.errorUtil as errorUtil

# Regular expressions:
MACRO_NAME_EXP = "[a-zA-Z0-9_\\@\\^\\<]"
MACRO_NAME_CHAR_REGEXP = re.compile(MACRO_NAME_EXP)
MACRO_SET_REGEXP = re.compile("\\s*([:+?]?)\\=\\s*")
IS_MACRO_DEF_REGEXP = re.compile("^%s+\\s*[:+?]?\\=.*" % MACRO_NAME_EXP, re.IGNORECASE)
IS_MACRO_INVOKE_REGEXP = re.compile(".*(?:[\\$])[\\(\\{]?%s+[\\)\\}]?" % MACRO_NAME_EXP)
SPACE_CHARS = re.compile("\\s")

# Constant(s)
COMMENT_CHAR = '#'

class MacroUtil:
    macroCommands = {} # All commands executable as $(name arg1, arg2, ...)
    definitionConditions = [] # A list of additional preconditions for the definition of a macro.
    lazyEvalConditions = []   # Don't expand macros on a line when in define & expand mode if any of these conditions are true.
    errorLogger = errorUtil.ErrorUtil()
    
    def setStopOnError(self, stopOnErr):
        self.errorLogger.setStopOnError(stopOnErr)
    def setSilent(self, silent):
        self.errorLogger.setSilent(silent)
    def setMacroCommands(self, commands):
        self.macroCommands = commands
    def addMacroDefCondition(self, condition):
        self.definitionConditions.append(condition)

    # Skip expanding macros on a line if condition holds. Applies only to
    # expandAndDefineMacros.
    def addLazyEvalCondition(self, condition):
        self.lazyEvalConditions.append(condition)

    # Get if [text] defines a macro.
    def isMacroDef(self, text):
        if IS_MACRO_DEF_REGEXP.match(text) == None:
            return False
        for condition in self.definitionConditions:
            if not condition(text):
                return False
        return True

    # Get whether [text] defines a macro with value that should be exported to the
    # environment.
    def isMacroExport(self, text):
        if not text.startswith("export "):
            return False
        return self.isMacroDef(text[len("export "):].strip())

    # Get if [text] syntatically invokes a macro.
    def isMacroInvoke(self, text):
        return IS_MACRO_INVOKE_REGEXP.match(text) != None

    # Get whether expandAndDefineMacros should
    # evaluate the contents of a line, or allow it to
    # be done later. Add conditions via addLazyEvalCondition.
    def shouldLazyEval(self, text):
        for condition in self.lazyEvalConditions:
            if condition(text):
                return True
        return False


    # Get a list of suggested default macros from the environment
    def getDefaultMacros(self):
        result = { }
        
        for name in os.environ:
            result[name] = os.environ[name]
        
        return result

    # Split content by lines, but
    # paying attention to escaped newline
    # characters.
    def getLines(self, content):
        result = []
        escapeCharLast = False
        buff = ''

        for c in content:
            if c == '\\' and not escapeCharLast:
                escapeCharLast = True
            elif escapeCharLast and c != '\n':
                buff += '\\' + c
                escapeCharLast = False
            elif escapeCharLast and c == '\n':
                buff += ' '
                escapeCharLast = False
            elif c == '\n':
                result.append(buff)
                buff = ''
            else:
                escapeCharLast = False
                buff += c

        result.append(buff)
        return result

    # Remove comments from line as defined
    # by COMMENT_CHAR
    def stripComments(self, line):
        singleLevel = { '"': False, "\'": False }
        inSomeSingleLevel = False
        multiLevelOpen = { '(': 0, '{': 0 }
        multiLevelClose = { ')': '(', '}': '{' }
        escaped = False
        trimToIndex = 0

        for c in line:
            if c in singleLevel and not escaped:
                if not inSomeSingleLevel:
                    inSomeSingleLevel = True
                    singleLevel[c] = True
                elif singleLevel[c]:
                    inSomeSingleLevel = False
                    singleLevel[c] = False
            elif c == '\\' and not escaped:
                escaped = True
            elif c == '\\' and escaped:
                escaped = False
            elif c in multiLevelOpen and not escaped and not inSomeSingleLevel:
                multiLevelOpen[c] += 1
            elif c in multiLevelClose and not escaped and not inSomeSingleLevel:
                bracketPairChar = multiLevelClose[c]
                if multiLevelOpen[bracketPairChar] == 0:
                    self.errorLogger.reportError("Parentheses mismatch on line with content: %s" % line)
                else:
                    multiLevelOpen[bracketPairChar] -= 1
            elif c == COMMENT_CHAR and not escaped and not inSomeSingleLevel:
                break
            else:
                escaped = False
            trimToIndex = trimToIndex + 1
        return line[:trimToIndex]

    # Expand usages of [macros] in [line]. Make no definitions and expand
    # regardless of lazyEvalConditions.
    def expandMacroUsages(self, line, macros):
        expanded = ''
        buff = ''
        afterBuff = ''
        parenLevel = 0
        inMacro = False
        buffFromMacro = False

        line += ' ' # Force any macros at the
                    # end of the line to expand.

        for c in line:
            if c == '$' and not inMacro and parenLevel == 0:
                expanded += buff
                buff = ''
                inMacro = True
            elif c == '$' and parenLevel == 0 and inMacro and buff == '$':
                inMacro = False
                expanded += '$'
            elif (c == '(' or c == '{') and inMacro:
                parenLevel += 1
            elif (c == ')' or c == '}') and inMacro:
                parenLevel -= 1

                if parenLevel == 0:
                    inMacro = False
                    buffFromMacro = True
            elif inMacro and parenLevel == 0 and not MACRO_NAME_CHAR_REGEXP.match(c):
                inMacro = False
                buffFromMacro = True
                afterBuff += c
            else:
                buff += c

            if buffFromMacro:
                buffFromMacro = False
                buff = buff.lstrip()
                words = SPACE_CHARS.split(buff)

                if buff in macros:
                    buff = macros[buff]
                elif words[0] in self.macroCommands:
                    buff = self.macroCommands[words[0]](" ".join(words[1:]), macros)
                else:
                    self.errorLogger.reportError("Undefined macro %s. Context: %s." % (buff, line))

                expanded += buff + afterBuff
#               print("Expanded to %s." % (buff + afterBuff))
                buff = ''
                afterBuff = ''
        
        if parenLevel > 0:
            self.errorLogger.reportError("Unclosed parenthesis: %s" % line)

        # Append buff, but ignore trailing space.
        expanded += buff[:len(buff) - 1] + afterBuff
        return expanded

    # Expand and handle macro definitions 
    # in [contents]. This includes removing end-of-line comments.
    def expandAndDefineMacros(self, contents, macros = {}):
        lines = self.getLines(contents)
        result = ''

        for line in lines:
            line = self.stripComments(line)
            exporting = self.isMacroExport(line)
            
            # If either a macro export, or a setting a macro's value, without an export...
            if self.isMacroDef(line) or exporting:
                if exporting:
                    line = line[len("export "):]
                
                parts = MACRO_SET_REGEXP.split(line)
                name = parts[0]
                definedTo = line[len(name):]
                definedTo = MACRO_SET_REGEXP.sub("", definedTo, count=1) # Remove the first set character.
                defineType = MACRO_SET_REGEXP.search(line).group(1) # E.g. :,+,? so we can do += or ?=
                name = name.strip()
                
                doNotDefine = False
                concatWith = ''
                deferExpand = False
                
                # ?=, so only define if undefined.
                if defineType == '?' and name in macros:
                    doNotDefine = True
                elif defineType == '+' and name in macros:
                    concatWith = macros[name]
                elif defineType == '':
                    deferExpand = True
                
                # Depending on the operator, we might not want to define the macro...
                if not doNotDefine:
                    if not deferExpand:
                        macros[name] = concatWith + self.expandMacroUsages(definedTo, macros).rstrip('\n')
                    else:
#                    print("Expansion defered: %s = %s" % (name, definedTo))
                        macros[name] = concatWith + definedTo.rstrip('\n')
                    
                    if exporting:
                        os.environ[name] = macros[name]
#            print("%s defined to %s" % (name, macros[name]))
            elif self.isMacroInvoke(line) and not self.shouldLazyEval(line):
                result += self.expandMacroUsages(line, macros)
            else:
                result += line
            result += '\n'

        return (result, macros)



#!/usr/bin/python

import cmd, os, sys
import almost_make.utils.shellUtil.shellUtil as shell
import almost_make.utils.shellUtil.runner as runner
import almost_make.utils.macroUtil as macroUtility
import almost_make.utils.shellUtil.escapeParser as escapeParser
from almost_make.utils.printUtil import *

from almost_make.utils.argsUtil import parseArgs
from almost_make.version import printVersion

# See https://docs.python.org/3/library/cmd.html
# TODO:
#  * Autocomplete using Cmd.completedefault.

class SimpleShell(cmd.Cmd):
    prompt = '$ '
    command = ''
    macroUtil = macroUtility.MacroUtil()
    
    def __init__(self, useBaseCommands=True, defaultFlags={}):
        self.updatePrompt()
        self.macros = self.macroUtil.getDefaultMacros()
        self.defaultFlags = defaultFlags
        self.state = runner.ShellState()
        
        if useBaseCommands:
        	self.macros["_CUSTOM_BASE_COMMANDS"] = True # Use custom base commands for testing.
        
        self.macroUtil.setStopOnError(False)

        
        cmd.Cmd.__init__(self)
    
    def updatePrompt(self):
        if "PS1" in os.environ:
            self.prompt = escapeParser.parseEscapes(os.environ["PS1"])
        else:
            self.prompt = '$ '
        
        if self.command != "":
            if "PS2" in os.environ:
                self.prompt = escapeParser.parseEscapes(os.environ["PS2"])
            else:
                self.prompt = ". "
    
    def precmd(self, line):
        return line
    
    def runCommand(self, command):
        try:
            result, self.macros = shell.evalScript(command, self.macroUtil, self.macros, self.defaultFlags, state=self.state)
            
            if result != 0:
                cprint("Warning:", FORMAT_COLORS["YELLOW"])
                print(" Command exited with non-zero exit status, %s." % str(result))
            return True
        except Exception as e:
            cprint("Error running %s:\n%s" % (command, str(e)), FORMAT_COLORS["RED"])
            print()
            return False

    def default(self, line):
        if line == "EOF":
            sys.exit(0)
    
        if line.strip().endswith('\\'):
            line = line.strip()
            
            self.command += line[0:len(line) - 1]
            self.updatePrompt()
            
            return
        else:
            self.command += line
        
        self.runCommand(self.command)
        self.command = ""
        
        self.updatePrompt()

ARG_MAPPINGS = \
{
    'h': 'help',
    'v': 'version',
    'B': 'without-builtins',
    'p': 'system-pipe'
}

JUST_FLAGS = \
{
    'help', 'version', 'without-builtins',
    'system-pipe'
}

def printHelp():
    cprint("Help: \n", FORMAT_COLORS['YELLOW'])
    cprint(" Summary: ", FORMAT_COLORS['YELLOW'])
    print("Run an interactive version of the shell built into almake. This is a POSIX-like shell. It is not POSIX-compliant.")
    cprint(" Usage: almake_shell [options] [files...]\n", FORMAT_COLORS['YELLOW'])
    print("  ...where each filename in [files...] is an optional file to interpret. If files are given, interpret them before opening the shell.")
    print("Options include:")
    cprint("    -h, --help", FORMAT_COLORS['GREEN'])
    print("\t Print this message.")
    cprint("    --version", FORMAT_COLORS['GREEN'])
    print("\t Print version and licensing information.")
    cprint("    -B, --without-builtins", FORMAT_COLORS['GREEN'])
    print("\t Do not (re)define built-in commands (like echo). By default, echo, ls, dir, pwd, and perhaps other commands," +
    " are defined and override any commands with the same name already present in the system.")
    cprint("    -p, --system-pipe", FORMAT_COLORS['GREEN'])
    print("\t Rather than attempting to pipe output between commands (e.g. in ls | grep foo), send piped portions of the input " +
        "to the system's shell.")

def main():
    args = parseArgs(sys.argv, ARG_MAPPINGS, strictlyFlags=JUST_FLAGS)
    
    if 'help' in args:
        printHelp()
        sys.exit(1)
    elif 'version' in args:
        printVersion()
        sys.exit(1)
    else:
        builtins = True
        flags = []

        if 'without-builtins' in args:
            builtins = False

        if 'system-pipe' in args:
            flags.append(runner.USE_SYSTEM_PIPE)

        shell = SimpleShell(builtins, flags)

        for script in args['default']:
            scriptFile = open(script, 'r')
            lines = scriptFile.readlines()
            scriptFile.close()
            
            for line in lines:
                result = shell.runCommand(line)

                if not result:
                    break
            
        shell.cmdloop()

if __name__ == "__main__":
    main()



#!/usr/bin/python3

# A very simple command-runner.
# See: https://www.gnu.org/software/bash/manual/bash.html#Shell-Expansions

import shlex, os, re
import subprocess

import almost_make.utils.shellUtil.globber as globber

PRECEDENCE_LIST = [ '||', '&&', ";", '|', '>', '2>&1', '&' ]
TWO_ARGUMENTS = { "||", "&&", ";", '|', '>' }
PIPE_OUT_PERMISSIONS = 0o660

SPACE_CHARS = re.compile("\\s")

SYSTEM_SHELL = "system-shell"
USE_SYSTEM_PIPE = "use-system-pipe"

class ShellState:
    def __init__(self):
        self.cwd = os.getcwd()

# Get the number of [char] at the beginning of [text].
def getStartingCount(text, char):
    pos = 0
    
    while pos < len(text) and text[pos] == char:
        pos += 1
    
    return pos

# Get the number of [char] at the end of [text].
def getEndingCount(text, char):
    pos = len(text) - 1
    count = 0

    while pos >= 0 and text[pos] == char:
        pos -= 1
        count += 1

    return count

# Remove empty strings from a string array.
def removeEmpty(array):
    result = []

    for val in array:
        if val != "":
            result.append(val)
    
    return result

# Get the number of pairs of parentheses that
# surround the entirety of [splitText].
def getParenCount(splitText):
    if len(splitText) == 0:
        return 0

    startingParenCount = getStartingCount(splitText[0], '(')
    endingParenCount = getEndingCount(splitText[len(splitText) - 1], ')')
    return min(startingParenCount, endingParenCount)

# Remove parentheses that surround the contents
# of the splitText array, [buff].
def unwrapParens(buff):
    if len(buff) < 2:
        return buff
    result = []
    parenCount = getParenCount(buff)

    result.append(re.sub(r"^[(]{" + str(parenCount) + r"}", "", buff[0]))

    for i in range(1, len(buff) - 1):
        result.append(buff[i])

    result.append(re.sub(r"[)]{" + str(parenCount) + r"}$", "", buff[len(buff) - 1]))

    return removeEmpty(result)

# Remove leading and trailing quotation marks from the given text.
# E.g. "a and b" -> a and b, "a and b' -> "a and b'. Remove at most 1.
def stripQuotes(text):
    if len(text) < 2:
        return text
    firstChar = text[0]
    lastChar = text[-1]
    
    if firstChar != lastChar:
        return text
    
    if firstChar in { '"', "'" }:
        return text[1 : len(text) - 1]
    return text

# Surround [text] with quotation marks,
# escaping any internal quotations with a backslash.
# Acts similar to shlex.quote in Python 3.7, but always adds
# quotation marks.
def quote(text, quoteChar="'"):
    result = ""
    escaped = False

    for char in text:
        if char == quoteChar and not escaped:
            result += "\\"
        elif char == '\\' and not escaped:
            escaped = True
        elif escaped:
            escaped = False
        result += char

    return quoteChar + result + quoteChar
# Return whether text begins and ends with the same quoting character,
# **and** the quoting is un-interupted. For example, "a, b, c" -> True,
# "a, b, c' -> False, "a, b," "c" -> False.
def isQuoted(text):
    if len(text) < 2:
        return False
    if text[0] != text[-1]: # Must start and end with the same character.
        return False
    if not text[0] in { '"', "'" }:
        return False
    
    startQuote = text[0]

    escaped = False

    text = text[1:-1] # Remove starting & ending...

    # If something would have broken us out of the quote...
    for char in text:
        if char == startQuote and not escaped:
            return False # ... then it isn't completely quoted.
        elif char == '\\' and not escaped:
            escaped = True
        elif escaped:
            escaped = False
    
    # If the ending quote would have been escaped, it isn't quoted.
    return not escaped

# Cluster elements in [splitText] based on precedence.
# E.g. ['a', '||', 'b', '&&', 'c', '-l'] -> [['a'], '||', [['b'], '&&', ['c', '-l']]].
def cluster(splitText, level=0):
    result = []
    buff = []
    
    if getParenCount(splitText) > 0:
        level = 0

    splitText = unwrapParens(splitText)
    
    if level >= len(PRECEDENCE_LIST):
        return splitText

    searchFor = PRECEDENCE_LIST[level]

    if not searchFor in TWO_ARGUMENTS:
        if len(splitText) == 0:
            return splitText
        
        if splitText[len(splitText) - 1] == searchFor:
            return [ cluster(splitText[:len(splitText) - 1], level + 1), searchFor ]
        return cluster(splitText, level + 1)

    parenLevel = 0
    addedChunks = 0

    for chunk in splitText:
        if chunk.startswith("("):
            parenLevel += getStartingCount(chunk, '(')
        elif chunk.startswith(")"):
            parenLevel -= getEndingCount(chunk, ')')

            if parenLevel == 0 and len(buff) > 0:
                buff.append(chunk)
                
                continue
        elif parenLevel == 0:
            if chunk == searchFor and addedChunks == 0:
                result.append(cluster(buff, level + 1))
                buff = []
                addedChunks += 1
                continue
        buff.append(chunk)
    if addedChunks > 0:
        result.append(searchFor)
        result.append(cluster(buff, level))
    else:
        result.extend(cluster(buff, level + 1))
    return result

# Collapse a clustered command back into a string.
def collapse(clustered):
    result = ""
    
    if len(clustered) == 0:
        return result
    elif type(clustered[0]) == str:
        for part in clustered:
            if SPACE_CHARS.search(part.strip()) != None and not isQuoted(part):
                result += " " + quote(part)
            else:
                result += " " + part
        return result.strip()
    
    result = " ".join([ collapse(elem) for elem in clustered ])
    
    return result.strip()

# Glob all arguments in args, excluding the first and 
# quoted arguments.
def globArgs(args, state):
    cwd = os.path.abspath(state.cwd or '.')
    result = []
    isFirst = True
    
    for arg in args:
        if not isQuoted(arg.strip()) and not isFirst:
            result.extend(globber.glob(arg, cwd))
        else:
            result.append(arg)
            isFirst = False
    
    return result

# Returns the exit status of the command specified by args
# (e.g. [ 'ls', '-la' ]). If the command is in [customCommands],
# however, run the custom command, rather than the system command.
# If 'system-shell' is in flags, run this portion of the command in the 
# system's shell. This is useful in systems that support, say,
# the '|' operator, but the file-descriptor piping method fails.
# At the time of this writing, this was the case with iOS's a-Shell.
# If [blocking] is false, do not block and return a Popen object.
def rawRun(args, customCommands={}, flags=[], stdin=None, stdout=None, stderr=None, blocking=True, state=ShellState()):
    if "2>&1" in flags:
        stderr = stdout

    returnOverride = False

    # Force non-blocking.
    if "&" in flags:
        blocking = False
        returnOverride = True
        
    if len(args) == 0:
        return 0
    
    sysShell = SYSTEM_SHELL in flags or None

    if not sysShell:
        args = globArgs(args, state)
        args = [ stripQuotes(arg) for arg in args ]

    command = args[0].strip()

    if command in customCommands:
        result = customCommands[command](args, flags, stdin, stdout, stderr, state)
        
        if type(result) == bool:
            if result:
                return 0
            return 1
        if result == None:
            return 0
        return result

    if blocking:
        proc = subprocess.run(args, stdin=stdin, stdout=stdout, stderr=stderr, shell=sysShell, close_fds=False, cwd=state.cwd)
        return proc.returncode
    else:
        result = subprocess.Popen(args, stdin=stdin, stdout=stdout, stderr=stderr, shell=sysShell, close_fds=False, cwd=state.cwd)

        if returnOverride:
            return 0
        else:
            return result

def evalCommand(orderedCommand, customCommands={}, flags=[], stdin=None, stdout=None, stderr=None, blocking=True, state=ShellState()):
    if len(orderedCommand) == 0:
        return False
    if type(orderedCommand[0]) == str:
        return rawRun(orderedCommand, customCommands, flags, stdin=stdin, stdout=stdout, stderr=stderr, blocking=blocking, state=state)
    elif len(orderedCommand) == 2:
        recurseFlags = flags.copy()
        flags.append(orderedCommand)

        return evalCommand(orderedCommand[0], customCommands, recurseFlags, stdin=stdin, stdout=stdout, stderr=stderr, state=state)
    elif len(orderedCommand) == 3:
        operator = orderedCommand[1]

        # If we are to use the system's built-in pipe, we need to stop here.
        # Collapse the input and run it.
        if (operator == '|' or operator == '>') and USE_SYSTEM_PIPE in flags:
            runFlags = flags.copy()
            runFlags.append(SYSTEM_SHELL) # Use the system's shell to interpret.
            
            return rawRun(collapse(orderedCommand), customCommands, runFlags, stdin=stdin, stdout=stdout, stderr=stderr, state=state)

        if operator == '|':
            fdIn, fdOut = os.pipe()

            left = evalCommand(orderedCommand[0], customCommands, flags, stdin=stdin, stdout=fdOut, stderr=stderr, blocking = False, state=state)
            os.close(fdOut)

            # Run right with given stdin, stdout.
            right = evalCommand(orderedCommand[2], customCommands, flags, stdin=fdIn, stdout=stdout, stderr=stderr, state=state)
            
            os.close(fdIn)

            return right
        elif operator == '>':
            outfd = os.open(os.path.abspath(" ".join(orderedCommand[2])), os.O_WRONLY | os.O_CREAT, mode=PIPE_OUT_PERMISSIONS)

            left = evalCommand(orderedCommand[0], customCommands, flags, stdin=stdin, stdout=outfd, stderr=stderr, state=state)

            os.close(outfd)

            return left
        elif operator == '||' or operator == '&&':
            left = evalCommand(orderedCommand[0], customCommands, flags, stdin=stdin, stdout=stdout, stderr=stderr, state=state)

            if (operator == '||' and left != 0) or (operator == '&&' and left == 0):
                right = evalCommand(orderedCommand[2], customCommands, flags, stdin=stdin, stdout=stdout, stderr=stderr, state=state)

                return right
            return left
        elif operator == ';':
            left = evalCommand(orderedCommand[0], customCommands, flags, stdin=stdin, stdout=stdout, stderr=stderr, state=state)
            right = evalCommand(orderedCommand[2], customCommands, flags, stdin=stdin, stdout=stdout, stderr=stderr, state=state)

            if left != 0:
                return left
            return right
        else:
            raise SyntaxError("Unknown separator, %s." % operator)
    else:
        raise SyntaxError("Too many parts to expression, %s" % str(orderedCommand))

# Like shlex.split with punctuation grouping, but groups punctuation intelligently.
def shSplit(text):
    escaped = False
    inQuote = None
    result = []
    buff = ""
    
    # Split by spaces and parentheses (and other punctuation chars...).
    for char in text:
        buff += char
        if char in { '"', "'" } and not escaped:
            if inQuote == char:
                inQuote = None
            elif inQuote == None:
                inQuote = char
        elif char == '\\' and not escaped:
            escaped = True
        elif escaped:
            escaped = False
        elif char in [ '(', ')', '|', '&', '>', ';', ' ', '\t', '\n' ] and inQuote == None:
            result.append(buff[:-1])
            result.append(char)
            buff = ""
    
    result.append(buff)
    
    buff = ''
    filtered = []
    
    
    # Remove spaces, clump parentheses.
    for part in result:
        part = part.strip()
        
        if part != '':
            # If the direction of parentheses switches...
            if part == '(' and ')' in buff  or  part == ')' and '(' in buff:
                filtered.append(buff)
                buff = ''
        
            if part.startswith('(') or part.endswith(')'):
                buff += part
            elif buff != "":
                filtered.append(buff)
                buff = ''
                filtered.append(part)
            else:
                filtered.append(part)
    if buff != '':
        filtered.append(buff)
    
    return filtered
                

# Run a filter on lex's split output list.
# E.g. Map [ ... '2', '>&', '1', ...] to
# [... '2>&1' ...].
def filterSplitList(splitString):
    result = []
    buff = []
    
    splitString = splitString.copy()
    splitString.reverse()

    for part in splitString:
        buff.append(part)

        if buff == ['|', '|']:
            result.append('||')
            buff = []
        elif buff == ['&', '&']:
            result.append('&&')
            buff = []
        elif buff == ['>', '>']:
            result.append('>>')
            buff = []
        elif part in [ '|', '&', '>' ]:
            result.extend(buff[:len(buff) - 1])
            buff = [ part ]
    
    result.extend(buff)
    result.reverse() # Reverse so that we have ['&', '&', '&'] -> ['&', '&&'], not ['&&', '&']
    
    splitString = result
    result = []
    buff = []
    
    # A second pass for parts that may share characters with those above.
    for part in splitString:
        buff.append(part)
        if buff == ['2', '>', '&', '1']:
            result.append('2>&1')
            buff = []
        elif part == '2':
            result.extend(buff[:len(buff) - 1])
            buff = [ part ]
    result.extend(buff)
    
    return result

# Run the POSIX-like shell command [commandString]. Define
# any additional commands through [customCommands].
def runCommand(commandString, customCommands = {}, flags = [], state=ShellState()):
    # Note: punctuation_chars=True causes shlex to cluster ();&| runs.
    #       For example, a && b -> ['a', '&&', 'b'], instead of ['a', '&', '&', 'b'].
    #       It also, however, clusters runs we don't want, like a &&& b -> ['a', '&&&', 'b'].
    #       This, however, split 'a/b/c' into ['a', '/', 'b', '/', 'c'], which we REALLY DON'T WANT.
    #       As such, we're using 'shSplit' here.
    portions = filterSplitList(shSplit(commandString))
    ordered = cluster(portions) # Convert ['a', '&&', 'b', '||', 'c'] into
                                #       [[['a'], '&&', ['b']], '||', ['c']]
    return evalCommand(ordered, customCommands, flags, state=state)

if __name__ == "__main__":
    # Run directly? Run tests!
    print("Testing runner.py...")
    
    def assertEql(left, right, description, alternates=None):
        if left != right or alternates and left in alternates:
            raise Exception("%s != %s (%s)" % (left, right, description)) # We WANT to see this. Don't catch.
    
    assertEql("1", "1", "This should always pass. A test of assertEql.")
    assertEql(['1', ['2', '3']], ['1', ['2', '3']], "We rely on this for our tests.")

    assertEql(cluster([]), [], "Empty cluster")
    assertEql(cluster(['a']), ['a'], "Identity cluster")
    assertEql(cluster([ "a", "||", "b"]), [['a'], '||', ['b']], "Simple cluster test.")
    assertEql(cluster([ "a", "&&", "b" ]), [['a'], '&&', ['b']], "Cluster with ands")
    assertEql(cluster([ "a", "&&", "b", "||", "c" ]), [[['a'], '&&', ['b']], '||', ['c']], "Precedence clustering")
    assertEql(cluster([ "a", "||", "b", "||", "c" ]), [["a"], "||", [["b"], "||", ["c"]]], "Parentheses")
    assertEql(cluster([ "(", "a", "||", "b", ")" ]), [["a"], "||", ["b"]], "Outside parentheses.")
    assertEql(cluster([ "a", "&" ]), [ ['a'], '&' ], "Single argument operator.")
    assertEql(cluster([ "b", "2>&1" ]), [ ['b'], '2>&1' ], "Pipe err to out")
    assertEql(cluster([ 'c', '|', 'd', '||', '(', 'f', '&&', 'g', '||', '(', 'h', '))']), 
            [[ ['c'], '|', ['d'] ], '||', [ [ ['f'], '&&', ['g'] ], '||', ['h'] ] ], "More complicated; nested parentheses")
    assertEql(cluster([ '((((((((((', 'a', '))))))))))' ]), ['a'], "Lots of parentheses")

    assertEql(filterSplitList([]), [], "Empty split list")
    assertEql(filterSplitList([ '2', '>', '&', '1' ]), [ '2>&1' ], "Split to error re-direction.")
    assertEql(filterSplitList([ '1', '|', '|', '2' ]), ['1', '||', '2'], "Cluster or.")
    assertEql(filterSplitList([ '1', '|', '|', '2', '&', '&', '&', '3', '>', '4' ]), ['1', '||', '2', '&', '&&', '3', '>', '4'], "More complicated clustering")

    assertEql(shSplit("123"), ["123"], "Simple split test")
    assertEql(shSplit("1|2|3"), ["1", "|", "2", "|", "3"], "Splitting on pipes")
    assertEql(shSplit("1||2&&3"), ["1", "|", "|", "2", "&", "&", "3"], "More complicated, no-space splitting.")
    assertEql(shSplit("1 || 2"), ["1", "|", "|", "2"], "Pipe splitting with spaces.")
    assertEql(shSplit("ls -la"), ["ls", "-la"], "Space-splitting")
    assertEql(shSplit("'ls -la'"), ["'ls -la'"], "Quoted space-(not)splitting")
    assertEql(shSplit("'ls' '-la'"), ["'ls'", "'-la'"], "Quoted parts splitting")
    assertEql(shSplit("1  \t\n > 2 &\n   & 1>2     "), ["1", ">", "2", "&", "&", "1", ">", "2"], "Lots of spaces!")
    assertEql(shSplit("1;2"), ["1", ";", "2"], "Does it correctly split on semi-colons?")
    assertEql(shSplit("1(2)"), ["1", "(", "2", ")"], "What about parentheses?")
    assertEql(shSplit("ls -la && (echo -ne foo\\n || (ps))"), ["ls", "-la", "&", "&", "(", "echo", "-ne", "foo\\n", "|", "|", "(", "ps", "))"], "Something like sh.")
    assertEql(shSplit("((( )( )))"), ["(((", ")", "(", ")))"], "Almost all parentheses!")

    assertEql(filterSplitList(shSplit("ls || ls")), ["ls", "||", "ls"], "Shlex replacement test 1!")
    assertEql(filterSplitList(shSplit("ls && ps")), ["ls", "&&", "ps"], "Shlex replacement test 2!")
    assertEql(filterSplitList(shSplit("ls; ps")), ["ls", ";", "ps"], "Shlex replacement test 3!")
    assertEql(filterSplitList(shSplit("ls; (ps 2>&1 | grep 'foo && not foo')")), ["ls", ";", "(", "ps", "2>&1", "|", "grep", "'foo && not foo'", ")"], "Shlex replacement test 4!")
    assertEql(filterSplitList(shSplit("\"ls; (ps 2>&1 | grep 'foo && not foo')\"")), ["\"ls; (ps 2>&1 | grep 'foo && not foo')\""], "Shlex replacement test 5!")
    assertEql(filterSplitList(shSplit("make CFLAGS=thing LDFLAGS= CC=cc GCC=g++ A= TEST=33")), 
        [ "make", "CFLAGS=thing", "LDFLAGS=", "CC=cc", "GCC=g++", "A=", "TEST=33" ],
        "Shlex replacement test 6!")
    assertEql(filterSplitList(shSplit('TEST_MACRO="Testing1234=:= := This **should ** work! "')),
        ['TEST_MACRO="Testing1234=:= := This **should ** work! "'], "Quoting that starts in the middle?")
    
    assertEql(collapse([ "1", '&&', '2']), '1 && 2', "Simple collapse test.")
    assertEql(collapse([ "a b", '&&', 'c']), "'a b' && c", "Spaces and quoting.", { "\"a b\" && c" })
    assertEql(collapse([ "a", "2>&1", '|', 'c' ]), "a 2>&1 | c", "Other separators.")
    
    assertEql(globArgs(['a test', 'of some things', 'that should work'], ShellState()), ['a test', 'of some things', 'that should work'], "Test identity globbing")
    
    assertEql(quote("singleWord"), "\'singleWord\'", "Single-word quoting")
    assertEql(quote("two words", '"'), "\"two words\"", "Two-word, double-quoting")
    assertEql(quote("two words"), "\'two words\'", "Two-word, single-quoting")
    assertEql(quote("""assertEql(quote("two words", "'"), "\'two words\'", "Two-word, single-quoting")""", "'"), 
        r''' 'assertEql(quote("two words", "\'"), "\'two words\'", "Two-word, single-quoting")' '''.strip(), 
        "Quoting the single-quoting test.")
    assertEql(quote("[left-[br]][acket[[quoting", '['), "[\\[left-\\[br]]\\[acket\\[\\[quoting[", "Quoting with left-bracket characters!")



#!/usr/bin/python3

from almost_make.utils.printUtil import *

class ErrorUtil:
    stopOnError = True
    silent = False
    # On error, report [message] depending on [SILENT] and [STOP_ON_ERROR]
    def reportError(self, message):
        if not self.silent or self.stopOnError:
            cprint(str(message) + "\n", "RED", file=sys.stderr)
        
        if self.stopOnError:
            print ("Stopping.")
            sys.exit(1)

    # Option-setting functions
    def setStopOnError(self, stopOnErr):
        self.stopOnError = stopOnErr

    def setSilent(self, silent):
        self.silent = silent 



#!/usr/bin/python3

# Parses very simple Makefiles.
# Useful Resources:
#  - Chris Wellons' "A Tutorial on Portable Makefiles". https://nullprogram.com/blog/2017/08/20/ Accessed August 22, 2020
#  - GNUMake: https://www.gnu.org/software/make/manual/make.html Accessed August 22, 2020
#  - BSDMake:  http://khmere.com/freebsd_book/html/ch01.html Accessed Aug 22 2020 

import re, sys, os, subprocess, time, threading
# from concurrent.futures import ThreadPoolExecutor # We are **not** using this because adding an 
#                                                   # executor to the queue when in an executed thread can cause deadlock! See 
#                                                   # https://docs.python.org/3/library/concurrent.futures.html#threadpoolexecutor

from almost_make.utils.printUtil import cprint
import almost_make.utils.macroUtil as macroUtility
import almost_make.utils.shellUtil.shellUtil as shellUtility
import almost_make.utils.shellUtil.runner as runner
import almost_make.utils.shellUtil.globber as globber
import almost_make.utils.errorUtil as errorUtility

# Regular expressions
SPACE_CHARS = re.compile(r'\s')
INCLUDE_DIRECTIVE_EXP = re.compile(r"^\s*(include|\.include|-include|sinclude)")

# Targets that are used by this parser/should be ignored.
MAGIC_TARGETS = \
{
    ".POSIX",
    ".SUFFIXES"
}

class MakeUtil:
    recipeStartChar = '\t'
    silent = False
    macroCommands = {}
    maxJobs = 1
    currentJobs = 1 # Number of currently running jobs...
    jobLock = threading.Lock()
    pending = {} # Set of pending jobs.
    justPrint = False # Print commands, without evaluating.

    def __init__(self):
        self.macroCommands["shell"] = lambda code, macros: os.popen(code).read().rstrip(' \n\r\t')

        self.errorUtil = errorUtility.ErrorUtil()
        self.macroUtil = macroUtility.MacroUtil()

        self.macroUtil.setMacroCommands(self.macroCommands)
        self.macroUtil.addMacroDefCondition(lambda line: not line.startswith(self.recipeStartChar))
        self.macroUtil.addLazyEvalCondition(lambda line: line.startswith(self.recipeStartChar))

    def setStopOnError(self, stopOnErr):
        self.macroUtil.setStopOnError(stopOnErr)
        self.errorUtil.setStopOnError(stopOnErr)

    def setSilent(self, silent):
        self.silent = silent
        self.macroUtil.setSilent(silent)
        self.errorUtil.setSilent(silent)
    
    def setJustPrint(self, justPrint):
        self.justPrint = justPrint

    # Set the maximum number of threads used to evaluate targets.
    # Note, however, that use of a recursive build-system may cause more than
    # this number of jobs to be used/created.
    def setMaxJobs(self, maxJobs):
        self.maxJobs = maxJobs

    # Get a tuple.
    # First item: a map from target names
    #   to tuples of (dependencies, action)
    # Second item: A list of the targets
    #   with recipies.
    def getTargetActions(self, content):
        lines = content.split('\n')
        lines.reverse()
        
        result = {}
        currentRecipe = []
        targetNames = []
        specialTargetNames = []

        for line in lines:
            if line.startswith(self.recipeStartChar):
                currentRecipe.append(line[len(self.recipeStartChar) : ]) 
                # Use len() in case we decide to 
                # be less compliant and make it 
                # more than a character.
            elif len(line.strip()) > 0:
                if not ':' in line:
                    if len(currentRecipe) > 0:
                        self.errorUtil.reportError("Pre-recipe line must contain separator! Line: %s" % line)
                    else:
                        continue
                sepIndex = line.index(':')
                allGenerates = SPACE_CHARS.split(line[:sepIndex].strip())
                preReqs = line[sepIndex + 1 :].strip()
                
                # Get the dependencies.
                dependsOn = SPACE_CHARS.split(preReqs)
                for generates in allGenerates:
                    currentDeps = []
                    currentDeps.extend(dependsOn)
                    if generates in result:
                        oldDeps, oldRecipe = result[generates]
                        currentDeps.extend(oldDeps)
                        oldRecipe.reverse()
                        currentRecipe.extend(oldRecipe)
                    
                    # Clean up & add to output.
                    outRecipe = [] + currentRecipe
                    outRecipe.reverse()
                    result[generates] = (currentDeps, outRecipe)

                    if generates.startswith('.'):
                        specialTargetNames.append(generates)
                    else:
                        targetNames.append(generates)
                currentRecipe = []
        # Move targets that start with a '.' to
        # the end...
        targetNames.reverse()
        targetNames.extend(specialTargetNames)
        return (result, targetNames)

    # Generate [target] if necessary. Returns
    # True if generated, False if not necessary.
    def satisfyDependencies(self, target, targets, macros):
        target = target.strip()
        if not target in targets:
            # Can we generate a recipe?
            for key in targets.keys():
    #            print("Checking target %s..." % key)
                if "%" in key:
                    sepIndex = key.index("%")
                    beforeContent = key[:sepIndex]
                    afterContent = key[sepIndex + 1 :]
                    if target.startswith(beforeContent) and target.endswith(afterContent):
                        deps, rules = targets[key]
                        newKey = target
                        newReplacement = newKey[sepIndex : len(newKey) - len(afterContent)]
                        deps = " ".join(deps)
                        deps = deps.split("%")
                        deps = newReplacement.join(deps)
                        deps = deps.split(" ")

                        targets[newKey] = (deps, rules)
                        break
                elif key.startswith(".") and "." in key[1:]:
                    shortKey = key[1:] # Remove the first '.'
                    parts = shortKey.split('.') # NOT a regex.
                    requires = '.' + parts[0].strip()
                    creates = '.' + parts[1].strip()
                    
                    # Don't evaluate... The user probably didn't intend for us to
                    # make a recipe from this.
                    if len(parts) > 2:
                        continue
                    
                    if not targets[".SUFFIXES"]:
                        continue
                    
                    validSuffixes,_ = targets[".SUFFIXES"]
                    
                    # Are these valid suffixes?
                    if not creates in validSuffixes \
                            or not requires in validSuffixes:
                        continue
                    
                    # Does it fit the current target?
                    if target.endswith(creates):
                        deps,rules = targets[key]
                        
                        newDeps = [ dep for dep in deps if dep != '' ]
                        withoutExtension = target[: - len(creates)]
                        
                        newDeps.append(withoutExtension + requires)
                        
                        targets[target] = (newDeps, rules)
                        break
        selfExists = os.path.exists(target)
        selfMTime = 0

        if not target in targets:
            if selfExists:
                return False
            else:
                self.errorUtil.reportError("No rule to make %s." % target)
                return True # If still running, the user wants us to exit successfully.
        runRecipe = False
        deps, commands = targets[target]
        
        if selfExists:
            selfMTime = os.path.getmtime(target)

        def isPhony(target):
            if not ".PHONY" in targets:
                return False
            
            phonyTargets,_ = targets['.PHONY']
            return target in phonyTargets or target in MAGIC_TARGETS
        selfPhony = isPhony(target)
        
        def needGenerate(other):
            return isPhony(other) \
                or not os.path.exists(other) \
                or selfMTime > os.path.getmtime(other) \
                or not selfExists \
                or selfPhony

        if isPhony(target):
            runRecipe = True

        if not runRecipe:
            if not selfExists:
                runRecipe = True
            else:
                for dep in deps:
                    if isPhony(dep) or \
                        not os.path.exists(dep) \
                        or os.path.getmtime(dep) >= selfMTime:
                            runRecipe = True
                            break
        # Generate each dependency, if necessary.
        if not runRecipe:
            return False
        
        pendingJobs = []

        for dep in deps:
    #        print("Checking dep %s; %s" % (dep, str(needGenerate(dep))))
            if dep.strip() != "" and needGenerate(dep):
                self.jobLock.acquire()
                if self.currentJobs < self.maxJobs and not dep in self.pending:
                    self.currentJobs += 1
                    self.jobLock.release()

                    self.pending[dep] = threading.Thread(target=self.satisfyDependencies, args=(dep, targets, macros))

                    pendingJobs.append(dep)
                else:
                    self.jobLock.release()
                    self.satisfyDependencies(dep, targets, macros)

        for job in pendingJobs:
            self.pending[job].start()

        # Wait for all pending jobs to complete.
        for job in pendingJobs:
            self.pending[job].join()
            self.pending[job] = None

            self.jobLock.acquire()
            self.currentJobs -= 1
            self.jobLock.release()

        # Here, we know that
        # (1) all dependencies are satisfied
        # (2) we need to run each command in recipe.
        # Define several macros the client will expect here:
        macros["@"] = target
        macros["^"] = " ".join(deps)
        if len(deps) >= 1:
            macros["<"] = deps[0]

        for command in commands:
            command = self.macroUtil.expandMacroUsages(command, macros).strip()
            if command.startswith("@"):
                command = command[1:]
            elif not self.silent:
                print(command)
            haltOnFail = not command.startswith("-")
            if command.startswith("-"):
                command = command[1:]
            try:
                status = 0
                
                if self.justPrint:
                    print(command)
                elif not "_BUILTIN_SHELL" in macros:
                    status = subprocess.run(command, shell=True, check=True).returncode
                else:
                    defaultFlags = []
                    
                    if "_SYSTEM_SHELL_PIPES" in macros:
                        defaultFlags.append(runner.USE_SYSTEM_PIPE)
                    
                    status,_ = shellUtility.evalScript(command, self.macroUtil, macros, defaultFlags = defaultFlags)
                
                if status != 0 and haltOnFail:
                    self.errorUtil.reportError("Command %s exited with non-zero exit status, %s." % (command, str(status)))
            except Exception as e:
                if haltOnFail: # e.g. -rm foo should be silent even if it cannot remove foo.
                    self.errorUtil.reportError("Unable to run command:\n    ``%s``. \n\n  Message:\n%s" % (command, str(e)))
        return True
    
    # Handle all .include and include directives.
    def handleIncludes(self, contents, macros):
        lines = self.macroUtil.getLines(contents)
        lines.reverse()

        newLines = []
        inRecipe = False

        for line in lines:
            if line.startswith(self.recipeStartChar):
                inRecipe = True
            elif inRecipe:
                inRecipe = False
            elif INCLUDE_DIRECTIVE_EXP.search(line) != None:
                line = self.macroUtil.expandMacroUsages(line, macros)
                
                parts = runner.shSplit(line)
                command = parts[0].strip()

                parts = runner.globArgs(parts, runner.ShellState()) # Glob all...
                parts = parts[1:] # Remove leading include...
                ignoreError = False

                # Safe including?
                if command.startswith('-') or command.startswith('s'):
                    ignoreError = True

                for fileName in parts:
                    fileName = runner.stripQuotes(fileName)

                    if not os.path.exists(fileName):
                        if ignoreError:
                            continue

                        self.errorUtil.reportError("File %s does not exist. Context: %s" % (fileName, line))
                        return (contents, macros)
                    
                    if not os.path.isfile(fileName):
                        if ignoreError:
                            continue
                            
                        self.errorUtil.reportError("%s is not a file! Context: %s" % (fileName, line))
                        return (contents, macros)

                    try:
                        with open(fileName, 'r') as file:
                            contents = file.read().split('\n')
                            contents.reverse() # We're reading in reverse, so write in reverse.

                            newLines.extend(contents)
                        continue
                    except IOError as ex:
                        if ignoreError:
                            continue
                        
                        self.errorUtil.reportError("Unable to open %s: %s. Context: %s" % (fileName, str(ex), line))
                        return (contents, macros)
            newLines.append(line)

        newLines.reverse()

        return self.macroUtil.expandAndDefineMacros("\n".join(newLines), macros)

    # Run commands specified to generate
    # dependencies of target by the contents
    # of the makefile given in contents.
    def runMakefile(self, contents, target = '', defaultMacros={ "MAKE": "almake" }, overrideMacros={}):
        contents, macros = self.macroUtil.expandAndDefineMacros(contents, defaultMacros)
        contents, macros = self.handleIncludes(contents, macros)
        targetRecipes, targets = self.getTargetActions(contents)

        if target == '' and len(targets) > 0:
            target = targets[0]

        # Fill override macros.
        for macroName in overrideMacros:
            macros[macroName] = overrideMacros[macroName]

        satisfied = self.satisfyDependencies(target, targetRecipes, macros)

        if not satisfied and not self.silent:
            print("Not hing to be done for target ``%s``." % target)
        
        return (satisfied, macros)



#!/usr/bin/python3

# Parse escape sequences.

# Replace all escape sequences in [text] 
# with their mappings in [escapes] (without the leading '\\', see default argument).
def parseEscapes(text, escapes={ '033': '\033', 't': '\t', 'r': '\r', 'n': '\n', 'b': '\b', 'a': '\a', '\\': '\\' }):
    result = ""
    buff = ""
    escaped = False
    for char in text:
        buff += char
        
        if escaped and buff in escapes:
            result += escapes[buff]
            buff = ''
            escaped = False
        elif char == '\\':
            result += buff[:len(buff) - 1]
            buff = ""
            escaped = True
    result += buff
    return result

# Tests.
if __name__ == "__main__":
    def assertEql(a, b):
        if a != b:
            raise Exception("%s != %s" % (str(a), str(b)))
    assertEql(parseEscapes("a"), "a")
    assertEql(parseEscapes(""), "")
    assertEql(parseEscapes("ab"), "ab")
    assertEql(parseEscapes("a\\n"), "a\n")
    assertEql(parseEscapes("\\n"), "\n")
    assertEql(parseEscapes("a\\nb\\nc"), "a\nb\nc")
    assertEql(parseEscapes("\\\\"), "\\")
    assertEql(parseEscapes("A test of this\\n", {}), "A test of thisn")
    assertEql(parseEscapes("A \\033[32m test! \\033[0m"), "A \033[32m test! \033[0m")
    assertEql(parseEscapes("\\033[32mabc\\033[0m"), "\033[32mabc\033[0m")



#!/usr/bin/python3

# Support for globbing paths, given a relevant directory.
# See: https://www.gnu.org/software/bash/manual/bash.html#Filename-Expansion
#
# Note that https://docs.python.org/3/library/glob.html may also be useful...
#           It seems, however, that as of Python 3.8, glob only supports globbing
#           with paths relative to the current directory (os.getcwd()), not our
#           custom [cwd] variable. As such, we are using 
#           https://docs.python.org/3.7/library/fnmatch.html#module-fnmatch
#           instead.
# To-dos:
#  * Recursive globbing support with **

import os, fnmatch, re

HAS_PATTERN = re.compile(r"([^\\]{2})*[\*\[\]]")

# Glob the contents of [text] if it appears to be a path using
# patterns supported by Python's fnmatch module.
def glob(text, cwd):
    # First, determine if the text needs to be/can be globbed.
    canGlob = False
    canSimplify = False

    escaped = False

    # Does it look like a path?
    for char in text:
        if char in { '"', "'" } and not escaped:
            return [ text ]
        elif char == '\\' and not escaped:
            escaped = True
        elif escaped:
            escaped = False
        elif char in { ' ' }:
            return [ text ]
        elif char in { '[', ']', '*' }:
            canGlob = True
        elif char in { '~', '/' }:
            canSimplify = True
    
    matchingPaths = []

    # Expand the path. Note that on Windows, normcase converts
    # its argument to a Windows-like path. On other operating systems,
    # it does nothing.
    if canSimplify:
        text = os.path.normcase(os.path.expanduser(text))
    
    if canGlob:
        parts = text.split(os.sep)
        fringe = [(0, cwd)]
        
        while len(fringe) > 0:
            depth, path = fringe.pop()
            matches = []

            for currentDepth in range(depth, len(parts)):
                part = parts[currentDepth]
                if HAS_PATTERN.search(part) == None:
                    path = os.path.join(path, part)

                    # If we are trying to match in a non-existent place, give up on 
                    # this path!
                    if not os.path.exists(path):
                        matches = []
                        break
                    else:
                        matches = [ path ]
                    continue
                
                # We can only list files in directories!
                if not os.path.isdir(path):
                    matches = []
                    break
                
                potentialMatches = os.listdir(path or '.')
                potentialMatches = [ os.path.join(path, match) for match in potentialMatches ] # listdir just has filenames... We need the leading [path]...
                part = os.path.join(path, part)

                matches = fnmatch.filter(potentialMatches, part)

                searchWithDepth = currentDepth + 1
                if part.endswith('**'): # If recursive-globbing, we need to start at the same depth for all sub-directories...
                    searchWithDepth = currentDepth
                
                fringe.extend(zip([searchWithDepth] * len(matches), matches))
            
            if os.path.isabs(text):
                matchingPaths.extend([ os.path.normpath(match) for match in matches ])
            else:
                matchingPaths.extend([ os.path.relpath(match, os.path.abspath(cwd)) for match in matches ])
    
    # If we didn't find any matches, just return the text we were given (perhaps with 
    # tilde-expansion).
    if len(matchingPaths) == 0:
        matchingPaths = [ text ]

    return matchingPaths

# Run directly? Run tests!
if __name__ == "__main__":
    testDir = os.path.dirname(__file__)

    if testDir != '':
        os.chdir(testDir)
    
    def canonicalizePath(path):
        return os.path.abspath(
            os.path.realpath(
                os.path.normcase(
                    os.path.normpath(
                        os.path.expanduser(path)
                    )
                )
            )
        )

    def assertEql(a, b, message):
        if a != b:
            raise Exception("%s != %s (%s)" % (str(a), str(b), message))
    
    def assertHas(collection, item, message):
        if collection == None:
            raise Exception("%s is None (%s)" % (str(collection), message))
        if not item in collection:
            raise Exception("not %s in %s holds. (%s)" % (str(item), str(collection), message))
    
    def assertHasPath(pathCollection, path, message):
        assertHas(pathCollection, os.path.normcase(os.path.normpath(path)), message)

    def assertPathListsEql(a, b, message):
        pathSetA = set()
        pathSetB = set()

        for path in a:
            pathSetA.add(canonicalizePath(path))
        
        for path in b:
            pathSetB.add(canonicalizePath(path))
        
        assertEql(pathSetA, pathSetB, message)

    assertEql(glob("willNotMatch", "."), ["willNotMatch"], "A glob that does nothing!")
    assertPathListsEql(glob("*.txt", "."), ["*.txt"], "A glob that does nothing (but is a valid glob)!")
    assertHas(glob("*.py", "."), "globber.py", "Finding this script with a glob!")
    assertHas(glob("globber.*", "."), "globber.py", "Finding this script with a different glob!")
    assertHas(glob("globber.*", os.path.abspath('.')), "globber.py", "Finding this script with a different glob (and differently-formatted path)!")
    assertHasPath(glob("../*.py", "."), "../printUtil.py", "Finding another script with a glob.")
    assertHasPath(glob("../../*.py", "."), "../../cli.py", "Finding yet another script with a glob.")
    assertHasPath(glob("../../*/*.py", "."), "../printUtil.py", "Finding yet another script with a glob.")
    assertHasPath(glob("../../*/shellUtil", "."), ".", "Finding a folder with a glob.")
    assertHasPath(glob("../../**/runner.py", "."), "runner.py", "Recursive globbing.")
    assertHasPath(glob("**/viaDot.*", "../../tests/"), "testMisc/includeTest/viaDot.mk", "More recursive globbing.")


import setuptools
from almost_make.version import VERSION_STRING

readmeFile = open("README.md", "r")
longDescription = readmeFile.read()
readmeFile.close()

setuptools.setup(
    name="almost_make",
    version=VERSION_STRING,
    author="Henry Heino",
    author_email="personalizedrefrigerator@gmail.com",
    description="A pure-Python makefile parser and executor.",
    long_description=longDescription,
    long_description_content_type="text/markdown",
    url="https://github.com/personalizedrefrigerator/AlmostMake",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha"
    ],
    entry_points={
        "console_scripts": [
            "almake = almost_make.cli:main",
            "almake_shell = almost_make.utils.shellUtil.interactiveShell:main"
        ]
    },
    python_requires=">=3.7"
)



import os, re, shlex

SPACE_CHARS = re.compile('\\s')

# Parse given arguments.
# args: The list of arguments given to the program (e.g. from sys.argv). Note that
# any default arguments (not immediately after  a key) are put into a list under the
# key defaultArgKey. If no default args, the list is empty. For example, 
# ['make'] -> {'default': []}. 
# If a given argument or its single-character representative is in [strictlyFlags], it is
# considered a flag -- non-argument text after it is associated with [defaultArgKey], rather
# than the argument. For example, if foo is in  strictlyFlags, then [ ... --foo thing ...]
# results in { ... 'foo': True, 'default': [... 'thing' ...] ... }.
def parseArgs(args, 
        mappings = 
        {
            'h': 'help'
        }, 
        defaultArgKey = 'default',
        excludeFilename = True,
        strictlyFlags={'help'}):
    result = { }
    singleChars = []
    lastArgText = None
    if excludeFilename:
        args = args[1:] # Omit the filename.
    result[defaultArgKey] = []

    for chunk in args:
        if len(chunk) == 0:
            continue    
        
        if chunk.startswith("--") and len(chunk) > 2:
            if lastArgText:
                result[lastArgText] = True
            lastArgText = chunk[2:]
        elif chunk.startswith("-") and len(chunk) > 1:
            singleChars.extend(chunk[1:])
            
            if lastArgText:
                result[lastArgText] = True
                lastArgText = None
            
            # Permits single-characters mapping to multi-char
            # flags **with values**.
            if chunk[-1] in mappings and not (mappings[chunk[-1]] in result):    
                lastArgText = mappings[chunk[-1]]
        elif lastArgText: # Assign to previous.
            result[lastArgText] = chunk
            lastArgText = None
        else: # Default argument.
            result[defaultArgKey].append(chunk)
        
        # If lastArgText is a flag -- it can't have a value associated with it,
        # clear it so we don't associate a chunk with it.
        if lastArgText and lastArgText in strictlyFlags:
            result[lastArgText] = True
            lastArgText = None
    if lastArgText:
        result[lastArgText] = True

    for char in singleChars:
        if char in mappings and not (mappings[char] in result):
            result[mappings[char]] = True
    
    return result

# Fill in an already-populated argument list
# with arguments defined in an environment variable.
# This variable should have name [envVariable].
# If [givenOverridesNew] is false, then arguments
# found in the variable override those given.
# Returns output as a new argument map. [mappings]
# is used to parse arguments in the environment.
def fillArgsFromEnv(argList, envVariable, mappings, defaultArgKey='default', givenOverridesNew=True):
    if not envVariable in os.environ:
        return argList
    
    # Get the argument mapping from the environment variable...
    envArgList = shlex.split(os.environ[envVariable])
    argsFromEnv = parseArgs(envArgList, mappings, defaultArgKey, excludeFilename = False)

    result = {}
    
    # Single-line ifs... Common in Lua... 
    # I don't think I've seen them in Python... Is this bad style?
    firstMap = not givenOverridesNew and argList or argsFromEnv
    secondMap = givenOverridesNew and argList or argsFromEnv
    
    for key in firstMap:
        result[key] = firstMap[key]
    
    for key in secondMap:
        result[key] = secondMap[key]
    
    return result

# Save the list of arguments specified by [argMap]
# in the environment variable, [envVariable].
# Question: Do we need to clean up after exiting???
def saveArgsInEnv(argMap, envVariable, doNotSave, defaultKey="default"):
    argString = ""
    
    for key in argMap:
        if key in doNotSave:
            continue
    
        prefix = "--"
        defTo = str(argMap[key])
        defTo = shlex.quote(defTo)
        
        if key == defaultKey: # If the default, we have a list...
            prefix = "" # Quote each individually.
            defTo = " ".join([ shlex.quote(val) for val in key ]) # default arg stores a list.
        elif len(key) == 1:
            prefix = "-"
        
        if argMap[key] == True:
            argString += prefix + key
        else:
            argString += prefix + key + " " + defTo
        
        argString += " "
    
    os.environ[envVariable] = argString
